// lzw_compression.cpp: implementation of the compression class.
//
//////////////////////////////////////////////////////////////////////
// Created by: Luria Israel
//
//  This project implements the use of a dynamic LZW compression.
//  This code is free for use, and can be inserted into any MFC project.
//////////////////////////////////////////////////////////////////////
#include "lzw/lzw_compression.h"

#include "utils/string_utils.h"
//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////
namespace lzw
{
    compression::compression (unsigned int heap_size_in_bytes, bool expanded) : _dictionary(heap_size_in_bytes, expanded)
    {
        reset();
    }

    void compression::reset()
    {
        _dictionary.reset();
        _saveddata = 0;
        _totalbits = 0;
        calculate_bit_size(_dictionary.size());
    }

    compression::~compression()
    {
    }

    // compress a string into the destination
    size_t compression::compress (const char* source, size_t length, char* destination, size_t limit)
    {
        size_t         written(0);
        unsigned int   prefix(0);
        unsigned int   last_prefix(not_in_dictionary);

        _saveddata = 0;
        _totalbits = 0;

        // compress the source
        for (size_t index(0); index < length; ++index)
        {
            char    current(source[index]);
            int     result(_dictionary.get_entry(prefix, current));

            // either add to the dictionary, or change the prefix to the one in the dictionary
            if (result == not_in_dictionary)
            {
                compress_data(destination, written, limit, prefix);
                last_prefix = _dictionary.size();
                calculate_bit_size(last_prefix);
                _dictionary.add_entry(prefix, current);
                prefix = current & 0xff;
            }
            else if (result == last_prefix)
            {
                compress_data(destination, written, limit, prefix);
                prefix = current & 0xff;
            }
            else
            {
                prefix = result;
            }
        }

        // compress the remaining information in the prefix into the destination file
        compress_data(destination, written, limit, prefix);

        if (written < limit)
        {
            if (_totalbits > 0)
            {
                destination[written] = char((_saveddata >> 24) & 0xff);
                ++written;
            }

            if (written < limit)
            {
                destination[written] = 0;
            }
        }

        return written;
    }

    // decompress a string from the source
    size_t compression::decompress (const char* source, size_t length, char* destination, size_t limit)
    {
        unsigned int    prefix(0);
        unsigned int    data;
        size_t          index(0);
        size_t          written(0);

        _saveddata = 0;
        _totalbits = 0;

        // Pull integers out of the stream until the 'done' code is found
        while ((data = decompress_data(source, index, length)) != _done)
        {
#ifdef LOG_COMPRESSION
            std::cout << "decompress data: " << data << " index: " << index << " length: " << length;
            std::cout << " saved data: " << to_binary(_saveddata) << " totalbits: " << _totalbits;
            std::cout << " maxbits: " << _maxbits <<  " dictionary.size: " << _dictionary.size() << std::endl;
#endif
            int current(_dictionary.write_bytes(destination, written, limit, data));

            if (current == -1) exit(-1);

            int result(_dictionary.get_entry(prefix, current));

            // any new combination is put into the dictionary
            if (result == not_in_dictionary)
            {
                _dictionary.add_entry(prefix, current);
                calculate_bit_size(_dictionary.size());
            }

            prefix = data;
        }

        if (written < limit)
        {
            destination[written] = 0;
        }

        return written;
    }

    // keep shifting saved data right, pulling the top 8 bits off and writing them
    void compression::compress_data (char* destination, size_t& index, size_t length, unsigned int data)
    {
#ifdef LOG_COMPRESSION
        std::cout << "compress_data << " << data;
        std::cout <<  " maxbits: " << _maxbits <<  " dictionary.size: " << _dictionary.size() << std::endl;
#endif
        //    Move the data you want to write few bits to the left
        //    and combine it with the already existing data in the buffer
        _saveddata |= int(data) << (32 - _maxbits - _totalbits);

        //    Add the new added number of bits to the total bits counter
        _totalbits += _maxbits;

        //    Check if it's possible to enter the data to the file
        //    (over and equal a byte of data)
        while ((_totalbits >= 8) && (index < length))
        {
#ifdef LOG_COMPRESSION_WRITES
            std::cout << "compress_data writes " << ((_saveddata >> 24) & 0xff) << " saved data: " << to_binary(_saveddata) << " index: " << index << " length: " << length << " totalbits: " << _totalbits << std::endl;
#endif
            destination[index] = char((_saveddata >> 24) & 0xff);
            ++index;
            _saveddata <<= 8;
            _totalbits -= 8;
        }
    }

    // the inverse
    unsigned int compression::decompress_data (const char* source, size_t& index, size_t length)
    {
        if (index < length)
        {
            while ((_totalbits <= 24) && (index < length))
            {
                _saveddata |= (source[index] & 0xff) << (24 - _totalbits);
                ++index;
                _totalbits += 8;
            }
        }
        else
        {
            // no remaining bits => done
            if (_saveddata == 0 && _totalbits <= 0)
            {
                return _done;
            }
        }

        // these two lines could be a roll and a mask in asm
        unsigned int data(_saveddata >> (32 - _maxbits));

        _saveddata <<= _maxbits;

        _totalbits -= _maxbits;

        return data;
    }

    void compression::calculate_bit_size (unsigned int value)
    {
        for (int counter(8); counter < 32; ++counter)
        {
            if ((value >> counter) == 0)
            {
                _maxbits = counter;
                _done = (1 << _maxbits) - 1;
                break;
            }
        }
    }

    void compression::dump_dictionary ()
    {
        _dictionary.dump();
    }

    void compression::metrics ()
    {
        _dictionary.metrics();
    }
}
